from datetime import datetime
from typing import List, Dict, Any, Optional, Tuple, Union

from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.belief.belief_service import BeliefService
from analogai.foundation.extensions.belief_query.condition import Condition
from analogai.foundation.extensions.belief_query.belief_criteria_matcher import BeliefCriteriaMatcher
from analogai.foundation.common.utils.similarity.similarity_service import SimilarityService
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.constraint_operator import ConstraintOperator
from analogai.foundation import config

CriterionConstraint = Union[Any, Tuple[Union[str, ConstraintOperator], Any], Condition]


class LearningsSearchBuilder:
    def __init__(
        self,
        belief_service: BeliefService,
        user_id: str,
        agent_id: str,
        similarity_service: Optional[SimilarityService] = None,
    ):
        self._belief_service = belief_service
        self._user_id = user_id
        self._agent_id = agent_id
        self._similarity_service = similarity_service
        self._modifier_strategy = config.BELIEF_QUERY_MODIFIER_STRATEGY
        self._main_clause: Dict[Criterion, Union[CriterionConstraint, List[CriterionConstraint]]] = {}
        self._or_clauses: List[Dict[Criterion, CriterionConstraint]] = []

    def exactModifierMatch(self) -> "LearningsSearchBuilder":
        self._modifier_strategy = "exact"
        return self

    def ignoreModifier(self) -> "LearningsSearchBuilder":
        self._modifier_strategy = "ignore"
        return self

    def _add_constraint(self, criterion: Criterion, constraint: CriterionConstraint) -> None:
        existing = self._main_clause.get(criterion)
        if existing is None:
            self._main_clause[criterion] = constraint
            return
        if type(existing) is list:
            existing.append(constraint)
        else:
            self._main_clause[criterion] = [existing, constraint]

    def where(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._main_clause[criterion] = value
        return self

    def whereMin(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._add_constraint(criterion, (ConstraintOperator.Min, value))
        return self

    def whereMax(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._add_constraint(criterion, (ConstraintOperator.Max, value))
        return self

    def whereNotNull(self, criterion: Criterion) -> "LearningsSearchBuilder":
        self._add_constraint(criterion, (ConstraintOperator.NotNull,))
        return self

    def whereNull(self, criterion: Criterion) -> "LearningsSearchBuilder":
        self._add_constraint(criterion, (ConstraintOperator.Null,))
        return self

    def whereGroup(self, criteria: Dict[Criterion, Any]) -> "LearningsSearchBuilder":
        self._main_clause = dict(criteria)
        return self

    def orWhere(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: value})
        return self

    def orWhereMin(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: (ConstraintOperator.Min, value)})
        return self

    def orWhereMax(self, criterion: Criterion, value: Any) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: (ConstraintOperator.Max, value)})
        return self

    def orWhereNotNull(self, criterion: Criterion) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: (ConstraintOperator.NotNull,)})
        return self

    def orWhereNull(self, criterion: Criterion) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: (ConstraintOperator.Null,)})
        return self

    def orWhereGroup(self, criteria: Dict[Criterion, Any]) -> "LearningsSearchBuilder":
        self._or_clauses.append(dict(criteria))
        return self

    def whereIn(self, criterion: Criterion, values: List[Any]) -> "LearningsSearchBuilder":
        self._main_clause[criterion] = list(values)
        return self

    def orWhereIn(self, criterion: Criterion, values: List[Any]) -> "LearningsSearchBuilder":
        self._or_clauses.append({criterion: list(values)})
        return self

    def apply(self, criteria: Dict[Criterion, Any]) -> "LearningsSearchBuilder":
        for criterion, value in criteria.items():
            self._main_clause[criterion] = value
        return self

    def exists(self) -> bool:
        return len(self._execute()) > 0

    def list(self) -> List[Belief]:
        return self._execute()

    def count(self) -> int:
        return len(self._execute())

    def _execute(self) -> List[Belief]:
        beliefs = self._prefetch_beliefs()
        confirmed = [b for b in beliefs if self._is_confirmed_learning(b)]
        return [belief for belief in confirmed if self._matches_criteria(belief)]

    def _prefetch_beliefs(self) -> List[Belief]:
        if self._or_clauses:
            return self._belief_service.get_all_for_agent(
                self._agent_id,
                user_id=self._user_id,
            )

        if self._deductions_enabled():
            return self._belief_service.get_all_for_agent(
                self._agent_id,
                user_id=self._user_id,
            )

        if self._has_multi_value_constraint(Criterion.SubjectCaption) or self._has_multi_value_constraint(
            Criterion.ComplementCaption
        ) or self._has_multi_value_constraint(Criterion.ObjectCaption):
            return self._belief_service.get_all_for_agent(
                self._agent_id,
                user_id=self._user_id,
            )

        subject_caption = self._extract_caption_for_criterion(Criterion.SubjectCaption)
        complement_caption = self._extract_caption_for_criterion(
            Criterion.ComplementCaption
        ) or self._extract_caption_for_criterion(Criterion.ObjectCaption)

        if subject_caption:
            beliefs = self._belief_service.search_by_subject_notion_expression(
                agent_id=self._agent_id,
                subject_notion_expression=NotionExpression(str(subject_caption)),
            )
            return beliefs or []

        if complement_caption:
            beliefs = self._belief_service.search_by_complement_notion_expression(
                agent_id=self._agent_id,
                complement_notion_expression=NotionExpression(str(complement_caption)),
            )
            return beliefs or []

        return self._belief_service.get_all_for_agent(
            self._agent_id,
            user_id=self._user_id,
        )

    def _has_multi_value_constraint(self, criterion: Criterion) -> bool:
        constraint = self._main_clause.get(criterion)
        if isinstance(constraint, list):
            return True
        return False

    def _deductions_enabled(self) -> bool:
        return bool(
            getattr(self._belief_service, "_categorical_deduction_service", None)
            or getattr(self._belief_service, "_conditional_deduction_service", None)
        )

    def _extract_caption_for_criterion(self, criterion: Criterion) -> Optional[Any]:
        if criterion in self._main_clause:
            value = self._extract_constraint_value(self._main_clause.get(criterion))
            if value is not None:
                return value
        for clause in self._or_clauses:
            if criterion in clause:
                value = self._extract_constraint_value(clause.get(criterion))
                if value is not None:
                    return value
        return None

    def _extract_constraint_value(self, constraint: Any) -> Optional[Any]:
        if constraint is None:
            return None
        if isinstance(constraint, list):
            for item in constraint:
                value = self._extract_constraint_value(item)
                if value is not None:
                    return value
            return None
        if isinstance(constraint, Condition):
            return constraint.value
        if isinstance(constraint, (tuple, list)):
            if len(constraint) > 1:
                return constraint[1]
            return None
        return constraint

    def _is_confirmed_learning(self, belief: Belief) -> bool:
        if not belief.statements:
            return False
        return len(belief.statements) > 0

    def _matches_criteria(self, belief: Belief) -> bool:
        if not self._main_clause and not self._or_clauses:
            return True
        main_clause = self._effective_clause(self._main_clause)
        if self._main_clause and not main_clause:
            return True
        if main_clause and self._matches_clause(belief, main_clause):
            return True
        for clause in self._or_clauses:
            effective = self._effective_clause(clause)
            if clause and not effective:
                return True
            if effective and self._matches_clause(belief, effective):
                return True
        return False

    def _effective_clause(self, clause: Dict[Criterion, Any]) -> Dict[Criterion, Any]:
        if self._modifier_strategy != "ignore":
            return clause
        modifier_criteria = {
            Criterion.Location,
            Criterion.Time,
            Criterion.DeonticModality,
            Criterion.Quantity,
        }
        return {
            criterion: constraint
            for criterion, constraint in clause.items()
            if criterion not in modifier_criteria
        }

    def _matches_clause(
        self,
        belief: Belief,
        clause: Dict[Criterion, Union[CriterionConstraint, List[CriterionConstraint]]],
    ) -> bool:
        return BeliefCriteriaMatcher.matches_clause(
            belief,
            clause,
            similarity_service=self._similarity_service,
        )



from typing import Optional, List
import math

from analogai.foundation import config
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.modules.notion.notion_repository import NotionRepository
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.infrastructure.storage.in_memory.in_memory_storage_adapter_impl import InMemoryStorageAdapterImpl
from analogai.foundation.infrastructure.mappers.notion_storage_mapper import NotionStorageMapper


class InMemoryNotionRepositoryImpl(NotionRepository):
    COLLECTION_NAME = "notions"

    def __init__(self, storage_adapter=None):
        if storage_adapter is None:
            self.storage_adapter = InMemoryStorageAdapterImpl()
        else:
            self.storage_adapter = storage_adapter
        self._embedding_service = None
        if config.USE_VECTOR_SEARCH:
            from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService
            self._embedding_service = EmbeddingService()

    def _cosine_similarity(self, vector_a: List[float], vector_b: List[float]) -> float:
        if not vector_a or not vector_b or len(vector_a) != len(vector_b):
            return 0.0
        dot = sum(a * b for a, b in zip(vector_a, vector_b))
        norm_a = math.sqrt(sum(a * a for a in vector_a))
        norm_b = math.sqrt(sum(b * b for b in vector_b))
        if norm_a == 0 or norm_b == 0:
            return 0.0
        return dot / (norm_a * norm_b)

    def create(self, notion: Notion) -> Notion:
        data = NotionStorageMapper.to_dict(notion)
        if config.USE_VECTOR_SEARCH and self._embedding_service:
            data["embedding"] = self._embedding_service.encode(notion.caption)
        stored_data = self.storage_adapter.store(self.COLLECTION_NAME, data)
        if not stored_data:
            raise ValueError(f"Failed to store notion {notion.id}")
        return notion

    def find(
        self,
        agent_id: str,
        expression: NotionExpression,
        context_ids: Optional[List[str]] = None,
        similarity_threshold: Optional[float] = None,
    ) -> Optional[Notion]:
        if not config.USE_VECTOR_SEARCH or not self._embedding_service:
            query = {"caption": expression.caption, "agent_id": agent_id}
            all_notions = self.storage_adapter.retrieve_by_attributes(self.COLLECTION_NAME, query)

            if not all_notions:
                return None

            notion_data = all_notions[0]
            notion = NotionStorageMapper.from_dict(notion_data)
            return notion

        candidates = self.storage_adapter.retrieve_by_attributes(
            self.COLLECTION_NAME,
            {"agent_id": agent_id},
        )
        if not candidates:
            return None

        threshold = (
            similarity_threshold
            if similarity_threshold is not None
            else config.SIMILARITY_CUTOFF
        )
        query_embedding = self._embedding_service.encode(expression.caption)
        best_match = None
        best_score = 0.0
        for notion_data in candidates:
            embedding = notion_data.get("embedding")
            if embedding is None:
                embedding = self._embedding_service.encode(notion_data.get("caption", ""))
            score = self._cosine_similarity(query_embedding, embedding)
            app_logger.debug(
                "Vector similarity notion match: query='%s' candidate='%s' score=%.4f threshold=%.4f",
                expression.caption,
                notion_data.get("caption", ""),
                score,
                threshold,
            )
            if score >= threshold and score > best_score:
                best_score = score
                best_match = notion_data

        if not best_match:
            return None

        app_logger.debug(
            "Vector notion match selected: query='%s' candidate='%s' score=%.4f",
            expression.caption,
            best_match.get("caption", ""),
            best_score,
        )
        return NotionStorageMapper.from_dict(best_match)

    def find_by_id(self, notion_id: str) -> Optional[Notion]:
        result_data = self.storage_adapter.retrieve_by_id(self.COLLECTION_NAME, notion_id)
        if result_data:
            notion = NotionStorageMapper.from_dict(result_data)
            return notion
        return None

    def find_all_by_agent(self, agent_id: str) -> List[Notion]:
        results = self.storage_adapter.retrieve_by_attributes(
            self.COLLECTION_NAME,
            {"agent_id": agent_id},
        )
        return [NotionStorageMapper.from_dict(d) for d in results]




import os
os.environ.setdefault("ENVIRONMENT", "development")
os.environ.setdefault("MONGODB_URI", "mongodb://localhost:27017")
os.environ.setdefault("MONGODB_DATABASE", "test")
os.environ.setdefault("NEO4J_URI", "bolt://localhost:7687")
os.environ.setdefault("NEO4J_USER", "neo4j")
os.environ.setdefault("NEO4J_PASSWORD", "password")
os.environ.setdefault("NEO4J_DATABASE", "neo4j")



from typing import Any, Optional

from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.constraint_operator import ConstraintOperator


class Condition:
    def __init__(
        self,
        criterion: Criterion,
        operator: ConstraintOperator,
        value: Optional[Any] = None,
    ):
        self.criterion = criterion
        self.operator = operator
        self.value = value



from typing import Optional
from fastapi import HTTPException, Request, status
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.modules.agent.agent_service import AgentService
from analogai.foundation.extensions.authentication.interfaces.authentication_factory_interface import (
    AuthenticationFactoryInterface,
)
from analogai.foundation.extensions.authorization.interfaces.authorization_factory_interface import (
    AuthorizationDependency,
    AuthorizationFactoryInterface,
)
from analogai.foundation.extensions.authorization.core import AuthorizationOperation, AuthorizationService
from analogai.foundation.extensions.authorization.models import AuthorizationContext
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.extensions.auth.authentication.api_token.api_token_authentication_service import (
    ApiTokenAuthenticationService,
)
from analogai.foundation.extensions.auth.authorization.authority_policy import (
    AuthorityPolicy,
    DefaultAuthorityPolicy,
)
from analogai.foundation.extensions.auth.authorization.error_mappers import AuthorizationHttpErrorMapper

class AuthorizationFactory(AuthorizationFactoryInterface):
    def __init__(self):
        self._authorization_service: Optional[AuthorizationService] = None
        self._authentication_factory: Optional[AuthenticationFactoryInterface] = None
        self._agent_service: Optional[AgentService] = None
        self._api_token_service: Optional[ApiTokenAuthenticationService] = None
        self._authority_policy: Optional[AuthorityPolicy] = None
        self._error_mapper: Optional[AuthorizationHttpErrorMapper] = None

    def create_authorization_dependency(
        self,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> AuthorizationDependency:
        authentication_dependency = self._get_authentication_factory().create_authentication_dependency()
        authorization_service = self._get_authorization_service()
        agent_service = self._get_agent_service()

        async def dependency(
            request: Request,
            agent_id: str,
            operation_override: AuthorizationOperation = operation,
        ) -> AuthorizationContext:
            user = await authentication_dependency(request)

            if getattr(request.state, "api_auth_type", None) in {"user", "system", "guest"}:
                return self._build_api_authorization_context(
                    request=request,
                    user=user,
                    agent_id=agent_id,
                    operation_override=operation_override,
                    authorization_service=authorization_service,
                    agent_service=agent_service,
                )

            try:
                agent = authorization_service.find_and_authorize_agent(
                    user,
                    agent_id,
                    operation_override,
                )
            except ValueError as error:
                raise self._get_error_mapper().map(error) from error

            authority = self._extract_authority(user, agent)
            app_logger.info(
                "Authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s",
                user.id if user else None,
                agent.id,
                user.email if user else None,
                authority,
            )
            return AuthorizationContext(
                user=user,
                agent=agent,
                authority=authority,
                operation=operation_override,
            )

        return dependency

    def create_api_authorization_dependency(
        self,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> AuthorizationDependency:
        authentication_dependency = (
            self._get_authentication_factory().create_api_token_authentication_dependency()
        )
        authorization_service = self._get_authorization_service()
        agent_service = self._get_agent_service()

        async def dependency(
            request: Request,
            agent_id: str,
            operation_override: AuthorizationOperation = operation,
        ) -> AuthorizationContext:
            user = await authentication_dependency(request)
            return self._build_api_authorization_context(
                request=request,
                user=user,
                agent_id=agent_id,
                operation_override=operation_override,
                authorization_service=authorization_service,
                agent_service=agent_service,
            )

        return dependency

    def _get_authorization_service(self) -> AuthorizationService:
        if not self._authorization_service:
            self._authorization_service = AuthorizationService(self._get_agent_service())
        return self._authorization_service

    def _get_agent_service(self) -> AgentService:
        if not self._agent_service:
            from analogai.foundation.extensions.authorization import authorization_factory as authz_factory

            repository = authz_factory.create_repository_strategy().get_agent_repository()
            self._agent_service = AgentService(repository)
        return self._agent_service

    def _get_authentication_factory(self) -> AuthenticationFactoryInterface:
        if not self._authentication_factory:
            from analogai.foundation.extensions.authorization import authorization_factory as authz_factory

            self._authentication_factory = authz_factory.AuthenticationFactory()
        return self._authentication_factory

    def _get_api_token_service(self) -> ApiTokenAuthenticationService:
        if not self._api_token_service:
            self._api_token_service = ApiTokenAuthenticationService()
        return self._api_token_service

    def _get_authority_policy(self) -> AuthorityPolicy:
        if not self._authority_policy:
            self._authority_policy = DefaultAuthorityPolicy()
        return self._authority_policy

    def _get_error_mapper(self) -> AuthorizationHttpErrorMapper:
        if not self._error_mapper:
            self._error_mapper = AuthorizationHttpErrorMapper.default()
        return self._error_mapper

    def _extract_authority(self, user: User, agent: Agent) -> float:
        return self._get_authority_policy().resolve_role(user, agent)

    def _build_api_authorization_context(
        self,
        request: Request,
        user: User,
        agent_id: str,
        operation_override: AuthorizationOperation,
        authorization_service: AuthorizationService,
        agent_service: AgentService,
    ) -> AuthorizationContext:
        token_agent_id = getattr(request.state, "api_agent_id", None)
        auth_type = getattr(request.state, "api_auth_type", "system")

        if token_agent_id and agent_id and token_agent_id != agent_id:
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail="Agent mismatch for API token",
            )

        resolved_agent_id = token_agent_id or agent_id
        if not resolved_agent_id:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail="Missing agent_id",
            )

        if auth_type == "user":
            try:
                agent = authorization_service.find_and_authorize_agent(
                    user,
                    resolved_agent_id,
                    operation_override,
                )
            except ValueError as error:
                raise self._get_error_mapper().map(error) from error

            authority = self._get_authority_policy().resolve_api(auth_type)
            app_logger.info(
                "API authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s auth_type=%s",
                user.id if user else None,
                agent.id,
                user.email if user else None,
                authority,
                auth_type,
            )
            return AuthorizationContext(
                user=user,
                agent=agent,
                authority=authority,
                operation=operation_override,
            )

        agent = agent_service.find(resolved_agent_id)
        if not agent:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail="Agent not found",
            )

        authority = self._get_authority_policy().resolve_api(auth_type)
        app_logger.info(
            "API authorization context resolved: user_id=%s agent_id=%s user_email=%s authority=%s auth_type=%s",
            user.id if user else None,
            agent.id,
            user.email if user else None,
            authority,
            auth_type,
        )
        return AuthorizationContext(
            user=user,
            agent=agent,
            authority=authority,
            operation=operation_override,
        )



from analogai.foundation.extensions.auth.authentication.authentication_factory import AuthenticationFactory
from analogai.foundation.extensions.authentication.providers.auth0 import Auth0JWTService
from analogai.foundation.setup.factories.repository_selector import create_repository_strategy

__all__ = ["AuthenticationFactory", "Auth0JWTService", "create_repository_strategy"]



from typing import Optional, Dict, Any
from analogai.foundation import config
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.categorical_deduction.categorical_deduction import CategoricalDeduction
from analogai.foundation.common.dtos.notion_expression import NotionExpression


class ExpressionFactory:
    def __init__(self):
        pass

    def _maybe_lemmatize_text(self, text: str) -> str:
        if not config.LEMMATIZATION_ENABLED:
            return text
        from analogai.foundation.common.utils.lemmatization import lemmatize_text
        return lemmatize_text(text)

    def _lemmatize_criteria(self, criteria: Dict[Criterion, Any]) -> Dict[Criterion, Any]:
        if not criteria:
            return {}
        normalized = dict(criteria)
        subject = normalized.get(Criterion.SubjectCaption)
        if subject is not None:
            normalized[Criterion.SubjectCaption] = self._maybe_lemmatize_text(str(subject))
        obj = normalized.get(Criterion.ObjectCaption)
        if obj is not None:
            normalized[Criterion.ObjectCaption] = self._maybe_lemmatize_text(str(obj))
        return normalized
    
    def create_notion_expression(self, caption: str, attributes=None) -> NotionExpression:
        lemmatized_caption = self._maybe_lemmatize_text(caption)
        return NotionExpression(
            caption=lemmatized_caption,
            children=[],
            context=[],
            attributes=attributes,
        )
    
    def propagate_notion_expression(self, notion: Notion) -> NotionExpression:
        lemmatized_caption = self._maybe_lemmatize_text(notion.caption)
        return NotionExpression(
            caption=lemmatized_caption,
            children=[],
            context=[],
            attributes=notion.attributes,
        )
    
    def propagate_belief_expression(self, belief: Belief) -> "BeliefExpression":
        from analogai.foundation.common.dtos.belief_expression import BeliefExpression
        subject_expression = self.propagate_notion_expression(belief.subject_notion)
        object_expression = self.propagate_notion_expression(belief.complement_notion)
        
        return BeliefExpression(
            subject_notion_expression=subject_expression,
            complement_notion_expression=object_expression,
            relation=belief.relation,
            probability=belief.probability,
            validity=belief.validity,
            modifier=belief.modifier,
        )
    
    def propagate_categorical_deduction_expression(self, inference: CategoricalDeduction) -> "CategoricalDeductionExpression":
        from analogai.foundation.common.dtos.categorical_deduction_expression import CategoricalDeductionExpression
        major_premise_expression = self.propagate_belief_expression(inference.major_premise)
        minor_premise_expression = self.propagate_belief_expression(inference.minor_premise)
        subject_expression = self.propagate_notion_expression(inference.subject_notion)
        object_expression = self.propagate_notion_expression(inference.complement_notion)
        
        return CategoricalDeductionExpression(
            major_premise_expression=major_premise_expression,
            minor_premise_expression=minor_premise_expression,
            subject_notion_expression=subject_expression,
            complement_notion_expression=object_expression,
            relation=inference.relation,
            probability=inference.probability,
            validity=inference.validity,
            modifier=inference.modifier
        )

    def propagate_hypothetical_statement_expression(self, hypothetical: "HypotheticalStatement") -> "HypotheticalStatementExpression":
        from analogai.foundation.common.dtos.hypothetical_statement_expression import HypotheticalStatementExpression
        from analogai.foundation.modules.conditionals.hypothetical_statement import HypotheticalStatement

        if not isinstance(hypothetical, HypotheticalStatement):
            raise TypeError("hypothetical must be a HypotheticalStatement")

        causes = [
            self._lemmatize_criteria(getattr(cause, "criteria", {}))
            for cause in (hypothetical.causes or [])
        ]
        effect = self._lemmatize_criteria(getattr(hypothetical.effect, "criteria", {}))

        return HypotheticalStatementExpression(
            causes=causes,
            effect=effect,
            probability=hypothetical.probability,
            validity=hypothetical.validity,
            effect_belief_id=hypothetical.effect_belief_id,
        )




from analogai.foundation.extensions.auth.authorization.authorization_factory import AuthorizationFactory
from analogai.foundation.extensions.auth.authentication.authentication_factory import AuthenticationFactory
from analogai.foundation.setup.factories.repository_selector import create_repository_strategy

__all__ = ["AuthorizationFactory", "AuthenticationFactory", "create_repository_strategy"]




from .condition import Condition
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.constraint_operator import ConstraintOperator
from analogai.foundation.core.value_objects.cause import Cause
from .traversable_notion import TraversableNotion

__all__ = [
    "Condition",
    "Criterion",
    "ConstraintOperator",
    "Cause",
    "TraversableNotion",
]



from dotenv import load_dotenv, find_dotenv
import os
import sys
from importlib.util import find_spec
from pathlib import Path
from analogai.foundation.common.enums.relationship_type_positivity_threshold import RelationshipTypePositivityThreshold


def _find_project_root(start: Path) -> Path | None:
    current = start
    while True:
        if (current / "pyproject.toml").exists() or (current / ".env").exists():
            return current
        if current.parent == current:
            return None
        current = current.parent


def _find_module_root(module_name: str) -> Path | None:
    spec = find_spec(module_name)
    if spec and spec.origin:
        return _find_project_root(Path(spec.origin).resolve())
    return None


def _ordered_env_roots() -> list[Path]:
    roots: list[Path] = []
    env_file = os.getenv("ENV_FILE")
    if env_file:
        env_path = Path(env_file).expanduser().resolve()
        roots.append(env_path.parent)

    argv0 = Path(sys.argv[0]).expanduser().resolve() if sys.argv and sys.argv[0] else None
    if argv0 and argv0.exists():
        entry_root = _find_project_root(argv0.parent)
        if entry_root:
            roots.append(entry_root)

    deepthink_root = _find_module_root("deepthink")
    if deepthink_root:
        roots.append(deepthink_root)

    foundation_root = _find_project_root(Path(__file__).resolve())
    if foundation_root:
        roots.append(foundation_root)

    seen: set[Path] = set()
    ordered: list[Path] = []
    for root in roots:
        if root not in seen:
            ordered.append(root)
            seen.add(root)
    return ordered


def _load_envs_in_order() -> None:
    for root in _ordered_env_roots():
        env_path = root / ".env"
        if env_path.exists():
            load_dotenv(env_path, override=False)


_load_envs_in_order()


def get_required_env(key: str) -> str:
    value = os.getenv(key)
    if value is None:
        raise ValueError(f"Required environment variable '{key}' is not set")
    return value


def get_optional_env(key: str, default: str) -> str:
    return os.getenv(key, default)


ENVIRONMENT = os.getenv('ENVIRONMENT', 'production')

if ENVIRONMENT not in ['development', 'production']:
    raise ValueError(
        "Environment variable ENVIRONMENT must be one of: development or production"
    )
print(f"Environment is set to {ENVIRONMENT} at Foundation")

DEBUG = os.getenv('DEBUG', 'true').lower() == 'true'
LEMMATIZATION_ENABLED = os.getenv('LEMMATIZATION_ENABLED', 'false').lower() == 'true'
SIMILARITY_CUTOFF = float(os.getenv('SIMILARITY_CUTOFF', '0.4'))
USE_VECTOR_SEARCH = os.getenv('USE_VECTOR_SEARCH', 'false').lower() == 'true'
print(f"USE_VECTOR_SEARCH is set to {USE_VECTOR_SEARCH} threshold is {SIMILARITY_CUTOFF}")

RELATIONSHIP_TYPE_POSITIVITY_THRESHOLD_NEGATIVE = float(
    os.getenv(
        'RELATIONSHIP_TYPE_POSITIVITY_THRESHOLD_NEGATIVE',
        str(RelationshipTypePositivityThreshold.NEGATIVE.value),
    )
)
RELATIONSHIP_TYPE_POSITIVITY_THRESHOLD_UNCERTAIN = float(
    os.getenv(
        'RELATIONSHIP_TYPE_POSITIVITY_THRESHOLD_UNCERTAIN',
        str(RelationshipTypePositivityThreshold.UNCERTAIN.value),
    )
)

BELIEF_QUERY_MODIFIER_STRATEGY = os.getenv('BELIEF_QUERY_MODIFIER_STRATEGY', 'exact').lower()
if BELIEF_QUERY_MODIFIER_STRATEGY not in ['exact', 'ignore']:
    raise ValueError("BELIEF_QUERY_MODIFIER_STRATEGY must be 'exact' or 'ignore'")

MONGODB_URI = os.getenv('MONGODB_URI')
MONGODB_DATABASE = os.getenv('MONGODB_DATABASE')

NEO4J_URI = os.getenv('NEO4J_URI')
NEO4J_USER = os.getenv('NEO4J_USER')
NEO4J_PASSWORD = os.getenv('NEO4J_PASSWORD')
NEO4J_DATABASE = os.getenv('NEO4J_DATABASE')
NEO4J_IS_ENCRYPTED = os.getenv('NEO4J_IS_ENCRYPTED', 'false')



import unittest

from analogai.foundation.config import ENVIRONMENT

class MongoDBEnvironmentCheckMixin:
    def setUp(self):
        super().setUp()
        if ENVIRONMENT == 'development':
            self.skipTest(
                "CRITICAL: MongoDB tests cannot be run in development environment! "
                "Set ENVIRONMENT=staging or ENVIRONMENT=production to run MongoDB tests."
            )




from typing import Optional
from datetime import datetime, timedelta
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.utils.time_utils.tense import Tense
from analogai.foundation.common.utils.time_utils.time_converter import TimeConverter
from analogai.foundation.common.utils.time_utils.relative_time_formatter import RelativeTimeFormatter
from analogai.foundation.common.utils.time_utils.timezone_utils import TimezoneUtils


class TimeFormatter:
    
    @staticmethod
    def has_time_data(time_obj: Optional[Time]) -> bool:
        return TimeConverter.has_time_data(time_obj)
    
    @staticmethod
    def determine_tense(time_obj: Optional[Time], reference_time: Optional[datetime] = None, timezone: Optional[str] = None) -> Tense:
        if not TimeFormatter.has_time_data(time_obj):
            return Tense.PRESENT
        
        reference_time = TimezoneUtils.get_timezone_aware_datetime(timezone, reference_time)
        
        time_date = TimeConverter.to_datetime(time_obj, reference_time)
        if time_date is None:
            return Tense.PRESENT
        
        if time_date.tzinfo is None:
            time_date = time_date.replace(tzinfo=reference_time.tzinfo)
        
        has_specific_time = time_obj.hour is not None or time_obj.minute is not None
        
        time_date_only = time_date.date()
        reference_date_only = reference_time.date()
        
        if time_date_only < reference_date_only:
            return Tense.PAST
        elif time_date_only > reference_date_only:
            return Tense.FUTURE
        else:
            if has_specific_time and time_date < reference_time:
                return Tense.PAST
            else:
                return Tense.PRESENT
    
    @staticmethod
    def format_natural_language(time_obj: Optional[Time], reference_time: Optional[datetime] = None, timezone: Optional[str] = None) -> str:
        if not TimeFormatter.has_time_data(time_obj):
            return ""
        
        reference_time = TimezoneUtils.get_timezone_aware_datetime(timezone, reference_time)
        
        time_date = TimeConverter.to_datetime(time_obj, reference_time)
        if time_date is None:
            return RelativeTimeFormatter._format_absolute(time_obj)
        
        if time_date.tzinfo is None:
            time_date = time_date.replace(tzinfo=reference_time.tzinfo)
        
        has_specific_time = time_obj.hour is not None or time_obj.minute is not None
        
        time_date_only = time_date.date()
        reference_date_only = reference_time.date()
        days_diff = (time_date_only - reference_date_only).days
        
        if not has_specific_time:
            if days_diff == 0:
                return "today"
            elif days_diff == -1:
                return "yesterday"
            elif days_diff == 1:
                return "tomorrow"
            else:
                delta = timedelta(days=abs(days_diff))
                if days_diff < 0:
                    return TimeFormatter._format_past_time(delta, time_date, reference_time, time_obj)
                else:
                    return TimeFormatter._format_future_time(delta, time_date, reference_time, time_obj)
        
        if days_diff == 0 and time_obj.hour is not None and time_obj.hour == reference_time.hour and time_obj.minute is None:
            return "now"
        
        delta = time_date - reference_time
        total_seconds = delta.total_seconds()
        
        if total_seconds > 0:
            return TimeFormatter._format_future_time(delta, time_date, reference_time, time_obj)
        elif total_seconds < 0:
            return TimeFormatter._format_past_time(abs(delta), time_date, reference_time, time_obj)
        else:
            return "today"
    
    @staticmethod
    def format_for_range(time_obj: Optional[Time], reference_time: Optional[datetime] = None, timezone: Optional[str] = None) -> str:
        reference_time = TimezoneUtils.get_timezone_aware_datetime(timezone, reference_time)
        return RelativeTimeFormatter.format_for_range(time_obj, reference_time)
    
    @staticmethod
    def _format_future_time(delta: timedelta, time_date: datetime, reference: datetime, time_obj: Time) -> str:
        time_date_only = time_date.date()
        reference_date_only = reference.date()
        days_diff = (time_date_only - reference_date_only).days
        
        has_specific_time = time_obj.hour is not None or time_obj.minute is not None
        total_seconds = delta.total_seconds()
        
        if days_diff == 1:
            if has_specific_time and total_seconds < 86400:
                if total_seconds < 3600:
                    minutes = int(total_seconds / 60)
                    if minutes == 1:
                        return "in 1 minute"
                    else:
                        return f"in {minutes} minutes"
                else:
                    hours = int(total_seconds / 3600)
                    remaining_seconds = total_seconds % 3600
                    minutes = int(remaining_seconds / 60)
                    
                    if hours > 0 and minutes > 0:
                        if hours == 1 and minutes == 1:
                            return "in 1 hour and 1 minute"
                        elif hours == 1:
                            return f"in 1 hour and {minutes} minutes"
                        elif minutes == 1:
                            return f"in {hours} hours and 1 minute"
                        else:
                            return f"in {hours} hours and {minutes} minutes"
                    elif hours > 0:
                        if hours == 1:
                            return "in 1 hour"
                        else:
                            return f"in {hours} hours"
            return "tomorrow"
        
        if total_seconds < 3600:
            minutes = int(total_seconds / 60)
            if minutes == 0:
                return "in less than a minute"
            elif minutes == 1:
                return "in 1 minute"
            else:
                return f"in {minutes} minutes"
        
        elif total_seconds < 86400:
            hours = int(total_seconds / 3600)
            remaining_seconds = total_seconds % 3600
            minutes = int(remaining_seconds / 60)
            
            if hours > 0 and minutes > 0:
                if hours == 1 and minutes == 1:
                    return "in 1 hour and 1 minute"
                elif hours == 1:
                    return f"in 1 hour and {minutes} minutes"
                elif minutes == 1:
                    return f"in {hours} hours and 1 minute"
                else:
                    return f"in {hours} hours and {minutes} minutes"
            elif hours > 0:
                if hours == 1:
                    return "in 1 hour"
                else:
                    return f"in {hours} hours"
            elif minutes > 0:
                if minutes == 1:
                    return "in 1 minute"
                else:
                    return f"in {minutes} minutes"
        elif days_diff < 7:
            return f"in {days_diff} days"
        elif days_diff < 14:
            return "next week"
        elif days_diff < 30:
            weeks = days_diff // 7
            return f"in {weeks} weeks"
        elif days_diff < 60:
            return "next month"
        elif days_diff < 365:
            months = days_diff // 30
            if months == 1:
                return "next month"
            return f"in {months} months"
        else:
            years = days_diff // 365
            if years == 1:
                return "next year"
            else:
                return f"in {years} years"
    
    @staticmethod
    def _format_past_time(delta: timedelta, time_date: datetime, reference: datetime, time_obj: Time) -> str:
        total_seconds = delta.total_seconds()
        
        if total_seconds < 3600:
            minutes = int(total_seconds / 60)
            if minutes == 0:
                return "just now"
            elif minutes == 1:
                return "1 minute ago"
            else:
                return f"{minutes} minutes ago"
        
        hours = int(total_seconds / 3600)
        
        time_date_only = time_date.date()
        reference_date_only = reference.date()
        
        if time_date_only == reference_date_only:
            if hours <= 6:
                if hours == 1:
                    return "1 hour ago"
                return f"{hours} hours ago"
            return "today"
        
        days_diff = (reference_date_only - time_date_only).days
        
        if days_diff == 1:
            return "yesterday"
        elif days_diff < 7:
            return f"{days_diff} days ago"
        elif days_diff < 14:
            return "last week"
        elif days_diff < 30:
            weeks = days_diff // 7
            return f"{weeks} weeks ago"
        elif days_diff < 60:
            return "last month"
        elif days_diff < 365:
            months = days_diff // 30
            if months == 1:
                return "last month"
            return f"{months} months ago"
        else:
            years = days_diff // 365
            if years == 1:
                return "last year"
            if years > 10:
                return str(time_date.year)
            return f"{years} years ago"
    
    @staticmethod
    def is_relative_date(time_obj: Optional[Time], reference_time: Optional[datetime] = None) -> bool:
        if not TimeFormatter.has_time_data(time_obj):
            return False
        
        if reference_time is None:
            reference_time = datetime.now()
        
        time_date = TimeConverter.to_datetime(time_obj, reference_time)
        if time_date is None:
            return False
        
        time_date_only = time_date.date()
        reference_date_only = reference_time.date()
        
        tomorrow = reference_date_only + timedelta(days=1)
        yesterday = reference_date_only - timedelta(days=1)
        
        return time_date_only == reference_date_only or time_date_only == tomorrow or time_date_only == yesterday



# from analogai.foundation.common.config import ENVIRONMENT

# def prevent_tests_in_production():
#     if ENVIRONMENT == 'production':
#         raise RuntimeError(
#             "CRITICAL: Tests cannot be run in production environment! "
#             "Set ENVIRONMENT=staging or ENVIRONMENT=development to run tests."
#         )

# prevent_tests_in_production()



import uuid
from typing import Optional, List

from analogai.foundation.core.value_objects.attribute import Attribute


class Notion:
    def __init__(
        self,
        user_id: str,
        agent_id: str,
        caption: str,
        id: Optional[str] = None,
        attributes: Optional[List[Attribute]] = None,
    ):
        self.id = id if id else str(uuid.uuid4())
        self.user_id = user_id
        self.agent_id = agent_id
        self.caption = caption
        self.attributes = attributes

    def __str__(self) -> str:
        if not self.attributes:
            return self.caption
        parts = []
        for attr in self.attributes:
            tag = getattr(attr, "tag", None)
            value = getattr(attr, "value", None)
            unit = getattr(attr, "unit", None)
            if value is None and hasattr(attr, "values"):
                values = getattr(attr, "values", None)
                if values is not None:
                    value = values

            if value is None:
                if tag:
                    parts.append(str(tag))
                continue

            if isinstance(value, list):
                value_str = ", ".join(str(v) for v in value)
            else:
                value_str = str(value)

            unit_str = f" {unit}" if unit else ""
            if tag:
                parts.append(f"{tag}={value_str}{unit_str}")
            else:
                parts.append(f"{value_str}{unit_str}")

        if not parts:
            return self.caption
        return f"{self.caption} ({', '.join(parts)})"

    def __repr__(self) -> str:
        return (
            "Notion(id={!r}, user_id={!r}, agent_id={!r}, caption={!r}, attributes={!r})"
        ).format(self.id, self.user_id, self.agent_id, self.caption, self.attributes)




from analogai.foundation.modules.belief.belief_service import BeliefService
from analogai.foundation.extensions.belief_query.search_builder import LearningsSearchBuilder
from analogai.foundation.common.utils.similarity.similarity_service import SimilarityService
from analogai.foundation import config
from typing import Optional


class BeliefQueryService:
    def __init__(
        self,
        belief_service: BeliefService,
        similarity_service: Optional[SimilarityService] = None,
    ):
        if similarity_service is None and config.USE_VECTOR_SEARCH:
            raise ValueError("similarity_service is required when USE_VECTOR_SEARCH is enabled")
        self._belief_service = belief_service
        self._similarity_service = similarity_service

    def search(self, user_id: str, agent_id: str) -> LearningsSearchBuilder:
        return LearningsSearchBuilder(
            belief_service=self._belief_service,
            user_id=user_id,
            agent_id=agent_id,
            similarity_service=self._similarity_service,
    )